"""Davis interface"""

import json
import pathlib
from datetime import datetime
from typing import Optional, List, Union

import h5py
import numpy as np

from piv2hdf import UserDefinedHDF5Operation
from piv2hdf.interface import PIVParameterInterface, PIVData
from . import parameter as davis_parameter
from .const import DAVIS_SOFTWARE, DEFAULT_DATASET_LONG_NAMES
from ..config import get_config
from ..interface import PIVFile
from ..pivview.const import IGNORE_ATTRS, DIM_NAMES
from ..time import create_recording_datetime_dataset

try:
    import lvpyio as lv
except ImportError:
    raise ImportError('Package "lvpyio" not installed which is needed to read Davis files')
try:
    import h5rdmtoolbox as h5tbx
except ImportError:
    raise ImportError('Package "h5rdmtoolbox" not installed which is needed to write HDF files from Davis files')


def _compute_physical_quantity(data, offset, slope):
    return data * slope + offset


class VC7File(PIVFile):
    """Davis vc7 aka buffer file interface class"""
    suffix = ".vc7"
    __parameter_cls__ = davis_parameter.DavisParameterFile

    def __init__(self,
                 filename: Union[str, pathlib.Path],
                 parameter: Union[None, PIVParameterInterface, pathlib.Path] = None,
                 user_defined_hdf5_operations: Optional[
                     Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]] = None,
                 **kwargs
                 ):
        self.filename = pathlib.Path(filename)
        assert self.filename.exists(), FileNotFoundError(f'File {self.filename} not found!')
        self._frame = None
        self._attrs = None
        self._variables = None
        self._root_attributes = None
        super().__init__(filename, parameter, user_defined_hdf5_operations=user_defined_hdf5_operations, **kwargs)

    def read(self, relative_time, **kwargs) -> PIVData:
        """Read data from file."""
        build_coord_datasets = kwargs.get("build_coord_datasets", True)

        buffer = lv.read_buffer(str(self.filename))

        frame = buffer[0]

        ny, nx = frame.shape
        scales = frame.scales
        x = np.array([scales.x.slope * i + scales.x.offset for i in range(nx)])
        y = np.array([scales.y.slope * j + scales.y.offset for j in range(ny)])
        z = np.array([scales.z.slope * i + scales.z.offset for i in range(1)])
        ix = np.arange(0, nx)
        iy = np.arange(0, ny)

        data_attrs = {
            "x": dict(unit="m", description=scales.x.description),
            "y": dict(unit="m", description=scales.y.description),
            "ix": dict(unit="", description="Index in x-direction"),
            "iy": dict(unit="", description="Index in y-direction"),
            "z": dict(unit="m", description=scales.z.description),
            "reltime": dict(unit="s", description="Recording time since start"),
        }

        data = {
            "ix": ix,
            "iy": iy,
            "x": x,
            "y": y,
            "z": float(z),
            "reltime": relative_time
        }
        for name, comp in frame.components.items():
            if name not in ("x", "y", "ix", "iy", "reltime"):
                data_attrs[name] = dict(
                    unit=comp.scale.unit,
                    description=comp.scale.description,
                    slope=comp.scale.slope,
                    offset=comp.scale.offset
                )
                data[name] = comp.planes[0]
        return PIVData(data, data_attrs, {})

    def to_hdf(self,
               hdf_filename: pathlib.Path,
               relative_time: float,
               recording_dtime: Union[datetime, List[datetime]],
               z: Union[str, float, None] = None,
               **kwargs) -> pathlib.Path:
        """Convert file to HDF5 format."""
        if hdf_filename is None:
            _hdf_filename = pathlib.Path.joinpath(self.filename.parent, f'{self.filename.stem}.hdf')
        else:
            _hdf_filename = hdf_filename

        pivdata = self.read(relative_time, **kwargs)
        ny, nx = pivdata.data['y'].size, pivdata.data['x'].size
        with h5py.File(_hdf_filename, "w") as main:
            main.attrs['software'] = json.dumps(DAVIS_SOFTWARE)
            main.attrs['title'] = 'piv snapshot data'

            if recording_dtime is not None:
                ds_rec_dtime = create_recording_datetime_dataset(main, recording_dtime, name='time')

            for i, cname in enumerate(('x', 'y', 'ix', 'iy')):
                ds = main.create_dataset(
                    name=cname,
                    shape=pivdata.data[cname].shape,
                    maxshape=pivdata.data[cname].shape,
                    chunks=pivdata.data[cname].shape,
                    data=pivdata.data[cname],
                    dtype=pivdata.data[cname].dtype,
                    compression=get_config('compression'),
                    compression_opts=get_config('compression_opts'))
                for k, v in pivdata.data_attrs[cname].items():
                    ds.attrs[k] = v
                ds.make_scale(DEFAULT_DATASET_LONG_NAMES[cname])

            for i, cname in enumerate(('z', 'reltime')):
                ds = main.create_dataset(cname, data=pivdata.data[cname])
                for k, v in pivdata.data_attrs[cname].items():
                    ds.attrs[k] = v
                ds.make_scale(DEFAULT_DATASET_LONG_NAMES[cname])

            # Data Arrays
            _shape = (ny, nx)
            for k, v in pivdata.data.items():
                if k not in DIM_NAMES:
                    ds = main.create_dataset(
                        name=k,
                        shape=_shape,
                        dtype=v.dtype,
                        maxshape=_shape,
                        chunks=_shape,
                        compression=get_config('compression'),
                        compression_opts=get_config('compression_opts'))
                    ds[:] = v
                    for ic, c in enumerate((('y', 'iy'), ('x', 'ix'))):
                        ds.dims[ic].attach_scale(main[c[0]])
                        ds.dims[ic].attach_scale(main[c[1]])
                    if recording_dtime is None:
                        ds.attrs['COORDINATES'] = ['reltime', 'z']
                    else:
                        ds.attrs['COORDINATES'] = ['reltime', 'z', ds_rec_dtime.name]

                    if k in pivdata.data_attrs:
                        for attr_key, attr_val in pivdata.data_attrs[k].items():
                            if attr_key not in IGNORE_ATTRS:
                                ds.attrs[attr_key] = attr_val

            # for name, comp in frame.components.items():
            #     ds = main.create_dataset(
            #         name=name,
            #         shape=comp.shape,
            #         maxshape=comp.shape,
            #         data=comp.planes[0],
            #         compression=get_config('compression'),
            #         compression_opts=get_config('compression_opts')
            #     )
            #     ds.attrs["description"] = comp.scale.description
            #     ds.attrs["unit"] = comp.scale.unit
            #     ds.attrs["slope"] = comp.scale.slope
            #     ds.attrs["offset"] = comp.scale.offset
            #
            #     for ic, c in enumerate((('y', 'iy'), ('x', 'ix'))):
            #         ds.dims[ic].attach_scale(main[c[0]])
            #         ds.dims[ic].attach_scale(main[c[1]])
            #
            #     if recording_dtime is None:
            #         ds.attrs['COORDINATES'] = ['reltime', 'z']
            #     else:
            #         ds.attrs['COORDINATES'] = ['reltime', 'z', ds_rec_dtime.name]
        return _hdf_filename



import warnings

from . import tools
from .file import OpenPIVFile, get_files
from .parameter import OpenPIVParameterFile

try:
    import openpiv

    try:
        openpiv_version = openpiv.__version__
    except AttributeError:
        import pkg_resources

        openpiv_version = pkg_resources.get_distribution('openpiv').version
    from packaging import version as version_pkg

    if version_pkg.Version(openpiv_version) < version_pkg.Version('0.25.0'):
        warnings.warn(f'OpenPIV version is too old ({openpiv_version}). Please update to version 0.25.0 or newer.',
                      UserWarning)
    # else:
    #     print('OpenPIV version is ok.')
except ImportError:
    pass



import json
import pathlib
import warnings
from datetime import datetime
from typing import Dict, Union, List, Optional

import h5rdmtoolbox as h5tbx
import numpy as np
import pandas as pd

from .const import *
from .parameter import OpenPIVParameterFile
from ..config import get_config
from ..flags import flag_translation_dict, Flags
from ..interface import PIVFile, UserDefinedHDF5Operation, PIVData
from ..utils import get_uint_type, parse_z

__this_dir__ = pathlib.Path(__file__)


class OpenPIVResultData(PIVFile):
    """OpenPIV Result data, not a file!"""
    suffix: str = None
    parameter = OpenPIVParameterFile
    _parameter: OpenPIVParameterFile = None

    def __init__(self, data: Dict,
                 parameter: Union[None, OpenPIVParameterFile, pathlib.Path] = None,
                 **kwargs):
        _tmp_filename = pathlib.Path('__tmp__.txt')
        with open(_tmp_filename, 'w'):
            pass
        super().__init__(_tmp_filename, parameter, **kwargs)
        pathlib.Path(_tmp_filename).unlink()
        self.data = data  # data like x,y,u,v,mask etc. as it directly comes from openpiv

    def read(self, recording_time: float, **kwargs):
        pass

    def to_hdf(self,
               hdf_filename: pathlib.Path,
               relative_time: float,
               recording_dtime: Union[datetime, List[datetime]],
               z: Union[float, None] = None,
               **kwargs) -> pathlib.Path:
        raise NotImplementedError("Not yet implemented")


class OpenPIVFile(PIVFile):
    """Open PIV File interface class

    Note, that .txt file from openPIV will not have all datasets, when ou use the openPIV.tools.save()
    method.
    """
    suffix: str = '.txt'
    __parameter_cls__ = OpenPIVParameterFile

    def __init__(self,
                 *args,
                 user_defined_hdf5_operations: Optional[
                     Union[UserDefinedHDF5Operation, List[UserDefinedHDF5Operation]]] = None,
                 **kwargs):
        super().__init__(*args, user_defined_hdf5_operations=user_defined_hdf5_operations, **kwargs, )

    def read(self, relative_time: float, **kwargs) -> PIVData:
        """Read data from file."""
        px_mm_scale = float(self._parameter.param_dict['scaling_factor'])  # pixel/mm

        data = pd.read_csv(self.filename, delimiter='\t', na_values='     nan')

        _ix = data["# x"].to_numpy()
        _iy = data["y"].to_numpy()

        i = -1
        for i, x in enumerate(_ix[0:-1]):
            if (x - _ix[i + 1]) > 0:
                break
        ix = _ix[:i + 1]
        iy = _iy[::i + 1]
        nx = len(ix)
        ny = len(iy)
        x = ix / px_mm_scale  # pixel/mm * 1/pixel = mm
        y = iy / px_mm_scale  # pixel/mm * 1/pixel = mm

        data_dict = {k: v.to_numpy().reshape((ny, nx)) for k, v in data.items() if k not in ('# x', 'y')}
        data_dict['ix'] = ix.astype(get_uint_type(ix))
        data_dict['iy'] = iy.astype(get_uint_type(iy))
        data_dict['x'] = x
        data_dict['y'] = y
        data_dict['reltime'] = np.asarray(relative_time)

        variable_attributes = {'ix': dict(units='pixel'),
                               'iy': dict(units='pixel'),
                               'x': dict(units='m'),
                               'y': dict(units='m'),
                               'u': dict(units='m/s'),
                               'v': dict(units='m/s'),
                               'reltime': dict(units='s'),
                               'sig2noise': dict(units=''),
                               }
        flag_name = None
        if 'flags' in data_dict:
            flag_name = 'flags'
        elif 'flag' in data_dict:
            flag_name = 'flag'
        elif 'piv_flag' in data_dict:
            flag_name = 'piv_flag'
        elif 'piv_flags' in data_dict:
            flag_name = 'piv_flags'

        if flag_name:
            if 'openpiv' in flag_translation_dict:

                # assume no flags larger than 2**8
                data_dict[flag_name] = data_dict[flag_name].astype(np.uint8)

                # convert flags to package meaning
                _flag_bools = {k: data_dict[flag_name] == k for k in flag_translation_dict['openpiv']}
                for k, v in flag_translation_dict['openpiv'].items():
                    data_dict[flag_name][_flag_bools[k]] = v.value
            variable_attributes[flag_name] = dict(units=' ',
                                                  flag_meaning=json.dumps({flag.value: flag.name for flag in Flags}))

        if 'mask' in data_dict:
            data_dict[flag_name][data_dict.pop('mask').astype(bool)] = Flags.MASKED.value

        if 'w' in data_dict:
            variable_attributes['w'] = dict(units='m/s')

        return PIVData(data_dict,
                       variable_attributes,
                       {'software': json.dumps(OPENPIV_SOFTWARE)})

    def to_hdf(self,
               hdf_filename: pathlib.Path,
               relative_time: float,
               recording_dtime: Optional[Union[datetime, List[datetime]]] = None,
               z: Union[float, None] = None,
               **kwargs) -> pathlib.Path:
        """converts the snapshot into an HDF file"""
        pivdata = self.read(relative_time)
        if z is not None:
            pivdata.data['z'], pivdata.data_attrs['z']['units'] = parse_z(z)
        # building HDF file
        if hdf_filename is None:
            _hdf_filename = pathlib.Path.joinpath(self.filename.parent, f'{self.filename.stem}.hdf')
        else:
            _hdf_filename = hdf_filename

        with h5tbx.File(_hdf_filename, "w") as main:
            main.attrs['title'] = 'piv snapshot data'
            for ak, av in pivdata.root_attrs.items():
                main.attrs[ak] = av
            main.attrs['plane_directory'] = str(self.filename.parent.resolve())

            # # process piv_parameters. there must be a parameter file at the parent location
            # piv_param_grp = main.create_group(PIV_PARAMETER_GRP_NAME)
            #
            # self.write_parameters(piv_param_grp)

            # if recording_dtime is not None:
            #     ds_rec_dtime = create_recording_datetime_dataset(main, recording_dtime, name='time')
            var_list = []
            for varkey, vardata in pivdata.data.items():
                if vardata.ndim == 1:
                    ds = main.create_dataset(
                        name=varkey,
                        shape=vardata.shape,
                        maxshape=vardata.shape,
                        data=vardata,
                        compression=get_config('compression'),
                        compression_opts=get_config('compression_opts'),
                        attrs=pivdata.data_attrs.get(varkey, None))
                    ds.make_scale()
                elif vardata.ndim == 0:
                    main.create_dataset(name=varkey,
                                        data=vardata,
                                        dtype=vardata.dtype,
                                        attrs=pivdata.data_attrs[varkey])
                else:
                    ds = main.create_dataset(
                        name=varkey,
                        shape=vardata.shape,
                        chunks=vardata.shape,
                        maxshape=vardata.shape,
                        data=vardata,
                        dtype=vardata.dtype,
                        compression=get_config('compression'),
                        compression_opts=get_config('compression_opts'),
                        attrs=pivdata.data_attrs.get(varkey, None))
                    var_list.append(ds)
            for ds in var_list:
                ds.dims[0].attach_scale(main['y'])
                ds.dims[1].attach_scale(main['x'])
            for k, v in pivdata.data_attrs.items():
                if k in main:
                    for ak, av in v.items():
                        main[k].attrs[ak] = av
        return hdf_filename


def get_files(folder, suffix='.txt', parameter=None) -> List[OpenPIVFile]:
    """get OpenPIVFile instances from folder"""
    folder = pathlib.Path(folder)
    if not folder.is_dir():
        raise FileNotFoundError(f'{folder} is not a folder.')
    if not folder.exists():
        raise FileNotFoundError(f'{folder} does not exist.')
    openpiv_txt_filenames = sorted(folder.glob(f'*{suffix}'))
    folder = openpiv_txt_filenames[0].parent
    if parameter is None:
        parameter_filename_candiates = sorted(folder.glob('*.par'))
        if len(parameter_filename_candiates) > 1:
            warnings.warn(f'Multiple parameter files found in {folder}. Using {parameter_filename_candiates[0]}.')
        parameter = parameter_filename_candiates[0]
    if not isinstance(parameter, (str, pathlib.Path)):
        par = OpenPIVParameterFile.from_windef(parameter)
        par.save(folder / 'openpiv_parameters.par')
    return [OpenPIVFile(f) for f in openpiv_txt_filenames]



import datetime
import pathlib
import unittest

import h5py
import h5rdmtoolbox as h5tbx
import numpy as np

from piv2hdf import openpiv, pivview, tutorial
from piv2hdf.config import get_config, set_config
from piv2hdf.interface import PIVPlane
from piv2hdf.openpiv.user_operations import add_standard_name_operation as openpiv_add_standard_name_operation
from piv2hdf.pivview.user_operations import add_standard_name_operation as pivview_add_standard_name_operation
from piv2hdf.time import TimeVectorWarning

__this_dir__ = pathlib.Path(__file__).parent
CREATOR = {
    'name': 'Matthias Probst',
    'id': 'https://orcid.org/0000-0001-8729-0482',
    'role': 'Researcher'
}

try:
    import lvpyio as lv

    lvpyio_installed = True
except ImportError:
    lvpyio_installed = False


def validate_planar_plane_pivview(test, hdf_filename, nc_data, plane):
    """Test the hdf5 file created by pivview"""
    with h5tbx.File(hdf_filename) as h5:

        nt = h5['u'].shape[0]
        for k in ('time', 'reltime', 'image_index'):
            test.assertEqual(nt, h5[k].shape[0])

        test.assertEqual(np.datetime64(datetime.datetime(2023, 10, 1, 12)), h5.time[0].values)
        for i in range(1, nt):
            test.assertEqual(np.datetime64(datetime.datetime(2023, 10, 1, 12) + datetime.timedelta(seconds=i / 10)),
                             h5.time[i].values)

        for k in nc_data.keys():
            test.assertIn(k, h5)

        for i in range(h5['u'].shape[0]):
            nc_data = plane.list_of_piv_files[i].read(0).data
            for k in nc_data.keys():
                if k not in ('z', 'reltime', 'x', 'y', 'ix', 'iy', 'time', 'image_index'):
                    np.testing.assert_array_equal(h5[k][i, ...], nc_data[k][()])

        for k, v in h5.items():
            if isinstance(v, h5py.Dataset) and v.ndim == 3:
                test.assertEqual('/reltime', v.dims[0][0].name)
                test.assertEqual('/time', v.dims[0][1].name)
                test.assertEqual('/image_index', v.dims[0][2].name)
                test.assertEqual(1, v.dims[0][2][0])
                test.assertEqual('/y', v.dims[1][0].name)
                test.assertEqual('/iy', v.dims[1][1].name)
                test.assertEqual('/x', v.dims[2][0].name)
                test.assertEqual('/ix', v.dims[2][1].name)

        test.assertIn('software', h5.attrs)
        test.assertEqual(h5['x'].size, 31)
        test.assertEqual(h5['y'].size, 15)
        test.assertEqual(h5['z'].size, 1)
        test.assertAlmostEqual(h5['z'].values[()], 0.51, 6)
        test.assertEqual(h5['u'].shape, (8, 15, 31))
        test.assertEqual(h5['v'].shape, (8, 15, 31))
        test.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
        test.assertEqual('uint8', h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype)

        # test.assertEqual(h5['dudx'].attrs['long_name'], 'm/s')


class TestPlane(unittest.TestCase):

    def test_singleplane_stereo_pivview(self):
        from piv2hdf.pivview import PIVViewStereoNcFile

        stereo_plane_dir = __this_dir__ / "resources/pivview/stereo/plane_01"
        if stereo_plane_dir.exists():
            PIVViewStereoNcFile.__parameter_cls__(stereo_plane_dir / 'plane.cfg')

            plane_stereo = PIVPlane.from_folder(
                plane_directory=stereo_plane_dir,
                time_info=(datetime.datetime(2023, 10, 1, 12, 0, 0), 10),
                pivfile=PIVViewStereoNcFile,
                user_defined_hdf5_operations=pivview_add_standard_name_operation,
                parameter=stereo_plane_dir / "plane.cfg"
            )

            spiv_filename = plane_stereo.to_hdf(
                piv_attributes={'creator': CREATOR, 'piv_medium': 'air'})

            with h5tbx.File(spiv_filename) as h5:
                self.assertTrue('flag_meaning' in h5['piv_flags'].attrs)
                for gradient_name in ('dudx', 'dudy', 'dvdx', 'dvdy', 'dwdx', 'dwdy'):
                    self.assertTrue(gradient_name in h5)
                    self.assertEqual(f'velocity gradient {gradient_name[0:2]}/{gradient_name[2:]}',
                                     h5[gradient_name].attrs['long_name'])
                    self.assertEqual('1/s', h5[gradient_name].attrs['units'])
                    self.assertTrue('flag_meaning' not in h5[gradient_name].attrs)  # checking issue #4

    def test_singleplane_pivivew(self):
        set_config(postproc=[])
        self.assertEqual(get_config()['postproc'], [])

        plane_dir = tutorial.PIVview.get_plane_directory()
        plane = PIVPlane.from_folder(plane_directory=plane_dir,
                                     time_info=(datetime.datetime(2023, 10, 1, 12), 10),  # dtime and frequency
                                     pivfile=pivview.PIVViewNcFile,
                                     user_defined_hdf5_operations=pivview_add_standard_name_operation,
                                     prefix_pattern='*[0-9]')
        hdf_filename = plane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                        creator=CREATOR),
                                    z=0.51)
        pivdata = plane.list_of_piv_files[0].read(0)
        validate_planar_plane_pivview(self, hdf_filename, pivdata.data, plane)
        with h5tbx.File(hdf_filename) as h5:
            self.assertTrue('mean_u' not in h5)

        set_config(postproc=['compute_time_averages'])
        self.assertEqual(get_config()['postproc'], ['compute_time_averages'])
        hdf_filename = plane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                        creator=CREATOR),
                                    z=0.51)
        with h5tbx.File(hdf_filename) as h5:
            self.assertTrue('mean_u' in h5)

        with self.assertRaises(ValueError):
            set_config(postproc=['invalid'])

        time_vector = [datetime.datetime(2023, 10, 1, 12) + datetime.timedelta(seconds=i / 10) for i in range(100)]
        # now have a longer time vector:
        plane = PIVPlane.from_folder(plane_directory=plane_dir,
                                     time_info=time_vector[0:len(plane.list_of_piv_files)],
                                     pivfile=pivview.PIVViewNcFile,
                                     user_defined_hdf5_operations=pivview_add_standard_name_operation,
                                     prefix_pattern='*[0-9]')
        hdf_filename = plane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                        creator=CREATOR),
                                    z=0.51)
        validate_planar_plane_pivview(self, hdf_filename, pivdata.data, plane)

        # now have a longer time vector that is too long!
        with self.assertWarns(TimeVectorWarning):
            plane = PIVPlane.from_folder(plane_directory=plane_dir,
                                         time_info=time_vector[:],
                                         pivfile=pivview.PIVViewNcFile,
                                         user_defined_hdf5_operations=pivview_add_standard_name_operation,
                                         prefix_pattern='*[0-9]')
        hdf_filename = plane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                        creator=CREATOR),
                                    z=0.51)
        validate_planar_plane_pivview(self, hdf_filename, pivdata.data, plane)

    if lvpyio_installed:
        def test_singleplane_davis(self):
            from piv2hdf.davis import VC7File
            from piv2hdf.davis.parameter import DavisParameterFile

            vc7_filenames = tutorial.Davis.get_vc7_files()
            param = DavisParameterFile(vc7_filenames[0])
            plane_dir = vc7_filenames[0].parent
            plane = PIVPlane.from_folder(
                plane_directory=plane_dir,
                time_info=(datetime.datetime(2023, 10, 1, 12), 10),  # dtime and frequency
                pivfile=VC7File,
                parameter=param,
                # user_defined_hdf5_operations=pivview_add_standard_name_operation,
                prefix_pattern='*[0-9]')
            hdf_filename = plane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                            creator=CREATOR),
                                        z=0.51)

    def test_singleplane_pivivew_explicit_parameter(self):
        # explicitly pass parameter
        plane_dir = tutorial.PIVview.get_plane_directory()
        par_filename = plane_dir / 'piv_parameters.par'
        par = pivview.PIVviewParamFile(par_filename)
        plane_dir = tutorial.PIVview.get_plane_directory()
        plane = PIVPlane.from_folder(plane_dir,
                                     time_info=(datetime.datetime(2023, 10, 1, 12, 0, 0), 10),
                                     pivfile=pivview.PIVViewNcFile,
                                     user_defined_hdf5_operations=pivview_add_standard_name_operation,
                                     parameter=par)

        hdf_filename = plane.to_hdf(piv_attributes=dict(piv_medium='air',
                                                        creator=CREATOR),
                                    z=0.51)
        with h5tbx.File(hdf_filename) as h5:
            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 3:
                    self.assertEqual('/reltime', v.dims[0][0].name)
                    self.assertEqual('/time', v.dims[0][1].name)
                    self.assertEqual('/image_index', v.dims[0][2].name)
                    self.assertEqual(1, v.dims[0][2][0])
                    self.assertEqual('/y', v.dims[1][0].name)
                    self.assertEqual('/iy', v.dims[1][1].name)
                    self.assertEqual('/x', v.dims[2][0].name)
                    self.assertEqual('/ix', v.dims[2][1].name)
            self.assertIn('software', h5.attrs)
            self.assertEqual(h5['ix'].dtype, np.uint8)
            self.assertEqual(h5['iy'].dtype, np.uint8)
            self.assertEqual(h5['x'].size, 31)
            self.assertEqual(h5['x'].dtype, get_config()['dtypes']['x'])
            self.assertEqual(h5['y'].size, 15)
            self.assertEqual(h5['z'].size, 1)
            self.assertAlmostEqual(h5['z'].values[()], 0.51, 6)
            self.assertEqual(h5['u'].shape, (8, 15, 31))
            self.assertEqual(h5['v'].shape, (8, 15, 31))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype, 'uint8')

        plane_dir = tutorial.PIVview.get_plane_directory()
        par_filename = plane_dir / 'piv_parameters.par'
        par = pivview.PIVviewParamFile(par_filename)
        plane_dir = tutorial.PIVview.get_plane_directory()
        plane = PIVPlane.from_folder(plane_dir,
                                     time_info=(datetime.datetime(2023, 10, 1, 12, 0, 0), 10),
                                     pivfile=pivview.PIVViewNcFile,
                                     user_defined_hdf5_operations=pivview_add_standard_name_operation,
                                     parameter=par)

        hdf_filename = plane.to_hdf(
            piv_attributes=dict(piv_medium='air',
                                creator=CREATOR),
            z=0.51)
        self.assertEqual(hdf_filename, plane.hdf_filename)

        with h5tbx.File(hdf_filename) as h5:

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 3:
                    self.assertEqual('/reltime', v.dims[0][0].name)
                    self.assertEqual('/time', v.dims[0][1].name)
                    self.assertEqual('/image_index', v.dims[0][2].name)
                    self.assertEqual(1, v.dims[0][2][0])
                    self.assertEqual('/y', v.dims[1][0].name)
                    self.assertEqual('/iy', v.dims[1][1].name)
                    self.assertEqual('/x', v.dims[2][0].name)
                    self.assertEqual('/ix', v.dims[2][1].name)

            self.assertEqual(get_config()['dtypes']['x'], 'float32')
            self.assertEqual(h5['x'].dtype.type, np.float32)
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype, 'uint8')

    def test_singleplane_openpiv(self):
        plane_dir = tutorial.OpenPIV.get_plane_directory()
        plane = PIVPlane.from_folder(plane_dir,
                                     time_info=(datetime.datetime(2023, 10, 1, 12, 0, 0), 10),
                                     pivfile=openpiv.OpenPIVFile,
                                     user_defined_hdf5_operations=openpiv_add_standard_name_operation,
                                     )
        hdf_filename = plane.to_hdf(
            piv_attributes=dict(piv_medium='air', creator=CREATOR), z=0.51)
        with h5tbx.File(hdf_filename) as h5:

            for k, v in h5.items():
                if isinstance(v, h5py.Dataset) and v.ndim == 3:
                    self.assertEqual('/reltime', v.dims[0][0].name)
                    self.assertEqual('/time', v.dims[0][1].name)
                    self.assertEqual('/image_index', v.dims[0][2].name)
                    self.assertEqual(1, v.dims[0][2][0])
                    self.assertEqual('/y', v.dims[1][0].name)
                    self.assertEqual('/iy', v.dims[1][1].name)
                    self.assertEqual('/x', v.dims[2][0].name)
                    self.assertEqual('/ix', v.dims[2][1].name)

            self.assertIn('software', h5.attrs)
            self.assertEqual(h5['x'].size, 63)
            self.assertEqual(h5['y'].size, 59)
            self.assertEqual(h5['z'].size, 1)
            self.assertAlmostEqual(h5['z'].values[()], 0.51, 6)
            for k, ds in h5.items():
                if isinstance(ds, h5py.Dataset):
                    if k not in ('ix', 'iy', 'x', 'y', 'z', 'reltime', 'time', 'image_index'):
                        self.assertEqual(ds.shape, (4, 59, 63))
            self.assertIsNotNone(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}))
            self.assertEqual(h5.find_one({'standard_name': {'$regex': 'piv_flag'}}).dtype, 'uint8')



from typing import Dict

import h5rdmtoolbox as h5tbx

from piv2hdf import postproc

CONFIG = dict(
    time_unit='s',
    datetime_str='%Y-%m-%dT%H:%M:%SZ%z',
    attrs_unit_name='units',
    compression='gzip',
    compression_opts=5,
    convention=None,
    take_min_nt=True,  # False will fill datasets up with np.NA
    raise_layout_validation_error=True,
    dtypes={'x': 'float32',
            'y': 'float32',
            'z': 'float32',
            'reltime': 'float32',
            # ix, iy, iz, image_index are identified automatically (see utils.get_uint_type())!
            },
    postproc=['compute_time_averages',
              ],
    # post={'compute_significance': False,  # computes running mean, std and relative std
    #       'compute_velocity_magnitude': False,
    #       'compute_time_average': False,  # computes mean_u, mean_v, mean_vel_magnitude
    #       'compute_velocity_gradients': False,  # computes du/dx, du/dy, dv/dx, dv/dy
    #       'compute_vorticity': False},
)

PIV_ATTRS = dict()
_ORIG_PIV_ATTRS_KEYS = list(PIV_ATTRS.keys())


class _Setter:
    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        self._update(self.old)


class PIVAttrsSetter(_Setter):
    """Set the piv attributes. PIV attributes are written to the
    root group of the hdf file.

    Mandatory piv attributes are:
    - contact: str
        The orcid must be used, e.g. 'https://orcid.org/0000-0001-8729-0482'
    - piv_medium: str
        E.g. air, water, etc.
    """

    def __init__(self):
        self.old = {}

    def __call__(self, **kwargs):
        self.old = {}
        with h5tbx.use(get_config('convention')) as cv:
            print(f'Checking the PIV attributes based on convention currently set ({cv.name})...')
            root_standard_attributes = cv.properties.get(h5tbx.File, None)
            if root_standard_attributes is not None:
                invalids = {}
                for k, v in kwargs.items():
                    sa = root_standard_attributes.get(k, None)
                    if sa is not None:
                        try:
                            sa.validate(v, None)
                            # sa.validator.model_validate(dict(value=v))
                        except Exception as e:
                            if sa.default_value == sa.__class__.EMPTY:
                                # collect invalids
                                invalids[k] = (v, e)

                if invalids:
                    invalids_str = '\n - '.join([f'"{k}": {v[0]} due to "{v[1]}"' for k, v in invalids.items()])
                    raise ValueError(f'Based on the set convention "{cv.name}",'
                                     f' the following PIV parameters are invalid: {invalids_str}')

        self._update(kwargs)

    def _update(self, options_dict: Dict):
        """Update piv attributes."""
        PIV_ATTRS.update(options_dict)


set_pivattrs = PIVAttrsSetter()


def get_pivattrs() -> Dict:
    """Return the piv attributes."""
    return PIV_ATTRS


def reset_pivattrs():
    PIV_ATTRS.clear()
    return PIV_ATTRS


class set_config(_Setter):
    """Set the configuration parameters."""

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k not in CONFIG:
                raise KeyError(f'Not a configuration key: {k}')
            if k == 'postproc':
                if not isinstance(v, list):
                    raise TypeError(f'Expected a list for key "postproc", got {type(v)}')

                for pp_name in v:
                    if isinstance(pp_name, str):
                        if pp_name not in dir(postproc):
                            raise ValueError(f'Unknown postproc name: {pp_name}')
                    elif isinstance(pp_name, callable):
                        pass
                    else:
                        raise TypeError(f'Expected a string or callable for key "postproc", got {type(pp_name)}')

            self.old[k] = CONFIG[k]
        self._update(kwargs)

    def _update(self, options_dict: Dict):
        CONFIG.update(options_dict)


def get_config(key=None):
    """Return the configuration parameters."""
    if key is None:
        return CONFIG
    return CONFIG[key]



"""
U0, V0, U1, V1, U2, V2, U3, V3, ACTIVE_CHOICE, ENABLED, TS:Peak ratio, MASK, TS:Correlation value, TS:IntWin angle,
 TS:IntWin factor X, TS:IntWin factor Y, TS:IntWin type, TS:Particle size, TS:Uncertainty V, TS:Uncertainty Vx, TS:Uncertainty Vy
"""

from .file import VC7File
from .parameter import DavisParameterFile

__all__ = ["VC7File", "DavisParameterFile"]



import datetime
from typing import Union, List

import h5py
import h5rdmtoolbox as h5tbx
import numpy as np


class TimeVectorWarning(Warning):
    pass


def create_recording_datetime_dataset(
        h5,
        recording_dtime: Union[datetime.datetime, List[datetime.datetime]],
        name: str) -> h5py.Dataset:
    """creates a recording datetime dataset in the hdf5 file"""
    if isinstance(recording_dtime, np.ndarray):
        ds_rec_dtime = h5tbx.Group(h5).create_time_dataset(
            name,
            data=list(recording_dtime),
            time_format="iso"
        )
    else:
        ds_rec_dtime = h5tbx.Group(h5).create_time_dataset(
            name,
            data=recording_dtime,
            time_format="iso"
        )
    ds_rec_dtime.attrs['comment'] = 'Recording datetime in ISO 8601 format'
    ds_rec_dtime.attrs['long_name'] = 'Datetime vector of snapshots'
    ds_rec_dtime.attrs['units'] = ''
    return ds_rec_dtime



from .nc import PIVViewNcFile, PIVViewStereoNcFile
from .parameter import PIVviewParamFile

config = {
    'interpolation': False,
    'masking': 'slack',  # all disabled and masked data points are set to np.nan
    'z_source': 'coord_min',
    'datetime_str': '%Y-%m-%dT%H:%M:%SZ%z',
    'attrs_unit_name': 'units',
    'compression': 'gzip',
    'compression_opts': 5,
    'take_min_nt': True,  # False will fill datasets up with np.NA
    'nc_as_time_averages_source': False,
}



import enum


class PIV_METHOD(enum.Enum):
    """Available PIV Methods"""
    single_pass = 0
    multi_pass = 1
    multi_grid = 2


class PIV_PeakFitMethod(enum.Enum):
    """Available PIV Peak Fit Methods"""
    Gauss3pt = 0  # openpiv has it, too
    Gauss3x3 = 1
    Parabolic = 2  # openpiv has it, too
    Centroid = 3  # openpiv has it, too
    Whittaker = 4
    NonLinGauss = 5
    FitBlob = 6
    FitRect = 7
    CentroidBinarized = 6



import pathlib
from itertools import count
from typing import Union, Tuple

import numpy as np
import yaml
from pint import Unit, Quantity

from . import cache, user

_filecounter = count()
_dircounter = count()
time_dimensionality = Unit('s').dimensionality
length_dimensionality = Unit('m').dimensionality


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def make_bold(string):
    """make string bold"""
    return f"{bcolors.BOLD}{string}{bcolors.ENDC}"


def read_structXML_wrapper(xmlFile):
    """
    This quick helper loads an existing vtr and converts it to a vtk dataset
    to use in mayavi. Add it with pipeline.add_dataset.

    Parameters
    ----------
    xmlFile : str
        Input stl data to load.

    Returns
    -------
    data : vtkRectilinearGrid
        Loaded XML data file.
    """
    try:
        import vtk
    except ImportError:
        ImportError('Package "vtk" not installed. Please install it first.')
    reader = vtk.vtkXMLRectilinearGridReader()
    reader.SetFileName(xmlFile)
    reader.Update()
    data = reader.GetOutput()
    return data


def is_time(unit):
    """
    Returns true if unit is a time unit, e.g. 'ms' or 's'
    """
    if unit == 's':
        return True
    else:
        return unit.dimensionality == time_dimensionality


def is_length(unit):
    """
    Returns true if unit is a time unit, e.g. 'ms' or 's'
    """
    if unit == 'm':
        return True
    else:
        return unit.dimensionality == length_dimensionality


def generate_temporary_filename(prefix='tmp', suffix: str = '') -> pathlib.Path:
    """generates a temporary filename in user tmp file directory

    Parameters
    ----------
    prefix: str, optional='tmp'
        prefix string to put in front of name
    suffix: str, optional=''
        suffix (including '.')

    Returns
    -------
    tmp_filename: pathlib.Path
        The generated temporary filename
    """
    _filename = user.TEMPORARY_USER_DIRECTORY / f"{prefix}{next(_filecounter)}{suffix}"
    while _filename.exists():
        _filename = user.TEMPORARY_USER_DIRECTORY / f"{prefix}{next(_filecounter)}{suffix}"
    cache.tmp_filenames.append(_filename)
    return _filename


def generate_temporary_directory(prefix='tmp') -> pathlib.Path:
    """generates a temporary directory in user tmp file directory

    Parameters
    ----------
    prefix: str, optional='tmp'
        prefix string to put in front of name

    Returns
    -------
    tmp_filename: pathlib.Path
        The generated temporary filename
    """
    _dir = user.TEMPORARY_USER_DIRECTORY / f"{prefix}{next(_dircounter)}"
    while _dir.exists():
        _dir = user.TEMPORARY_USER_DIRECTORY / f"{prefix}{next(_dircounter)}"
    _dir.mkdir(parents=True)
    cache.tmp_dirnames.append(_dir)
    return _dir


def validate_directory(directory_path: pathlib.Path) -> pathlib.Path:
    """Raise an error if directory does not exist or passed
    path is not a directory"""
    directory_path = pathlib.Path(directory_path)
    if not directory_path.exists():
        raise NotADirectoryError(f'Directory not found: {directory_path}')
    if not directory_path.is_dir():
        raise NotADirectoryError(f'Not a directory {directory_path}')
    return directory_path


def _get_uint(max_val):
    """get uint type based on max value"""
    if max_val < 2 ** 8:
        return 'uint8'
    if max_val < 2 ** 16:
        return 'uint16'
    if max_val < 2 ** 32:
        return 'uint32'
    return 'uint64'


def get_uint_type(data: Union[np.ndarray, int, float]) -> str:
    """estimate uint dtype based on data"""
    if isinstance(data, (int, float)):
        return _get_uint(data)
    return _get_uint(data.max())


def parse_z(z: Union[str, int, float]) -> Tuple[Union[int, float], str]:
    """Parses coordinate, which can be a number or a string

    Parameters
    ----------
    z: Union[str, int, float]
        The z-coordinate input. If a number (int, float), then unit is assumed to be [m].
        If a string is provided, it is processed by `pint.Quantity`

    Returns
    -------
    mag, unit: Tuple[Union[int, float], str]
        Magnitude and unit derived based on the input.
    """
    if isinstance(z, str):
        z_quantity = Quantity(z)  # .to('m')
        mag = z_quantity.magnitude
        unit = str(z_quantity.units)
    elif isinstance(z, (int, float)):
        unit = 'm'
        mag = z
    else:
        raise TypeError(f"z coordinate data must be a number or a string but got {type(z)}")
    return mag, unit


# def create_dataset(*, h5, name, **kwargs):
#     dtype = kwargs.pop('dtype', DEFAULT_CONFIG['dtypes'].get(name, None))
#     return h5.create_dataset(name=name, dtype=dtype, **kwargs)


def read_translation_yaml_file(filename):
    with open(filename, 'r') as f:
        pivview_translation = yaml.safe_load(f)
    return pivview_translation


def write_final_interrogation_window_size_to_h5_group(h5grp, interrogation_window_size, dtype=None):
    if isinstance(interrogation_window_size, int):
        interrogation_window_size = (interrogation_window_size, interrogation_window_size)
    if dtype is None:
        dtype = 'int16'
    if not isinstance(interrogation_window_size, (list, tuple)):
        raise TypeError(
            f'interrogation_window_size must be int or list or tuple, got {type(interrogation_window_size)}')

    if not len(interrogation_window_size) in (2, 3):
        raise ValueError(f'interrogation_window_size must have length 2, got {len(interrogation_window_size)}')

    for prefix, iws in zip(('x', 'y'), interrogation_window_size):
        iwsds = h5grp.create_dataset(f'{prefix}_final_iw_size',
                                     data=iws,
                                     dtype=dtype)
        iwsds.attrs.update({
            # 'standard_name': f'{prefix}_final_interrogation_window_size',
            'long_name': f'the final interrogation window size in {prefix}-direction',
            'units': 'pixel'})


def write_final_interrogation_overlap_size_to_h5_group(h5grp, overlap_size, dtype=None):
    if isinstance(overlap_size, int):
        overlap_size = (overlap_size, overlap_size)
    if dtype is None:
        dtype = 'int16'
    if not isinstance(overlap_size, (list, tuple)):
        raise TypeError(f'overlap_size must be int or list or tuple, got {type(overlap_size)}')

    if not len(overlap_size) == 2:
        raise ValueError(f'overlap_size must have length 2, got {len(overlap_size)}')

    for prefix, iws in zip(('x', 'y'), overlap_size):
        iwods = h5grp.create_dataset(f'{prefix}_final_overlap',
                                     data=iws,
                                     dtype=dtype)
        iwods.attrs.update({
            # 'standard_name': f'{prefix}_final_interrogation_window_overlap_size',
            'long_name': f'the final interrogation window size in {prefix}-direction',
            'units': 'pixel'})



"""const for openpiv module"""
OPENPIV_SOFTWARE = dict(name='openpiv', version=999,
                        url='https://github.com/matthiasprobst/openpiv-python-ext',
                        description="Forked version of openpiv-python with improved I/O and other small improvements.")



import pathlib
import shutil
from typing import Tuple, List

from . import user
from .utils import generate_temporary_directory

__this_dir__ = pathlib.Path(__file__).parent


class PIVview:
    """PIVview tutorial class"""

    @staticmethod
    def get_parameter_file() -> pathlib.Path:
        """Return pivview parameter file"""
        return user.TEST_DATA_DIRECTORY / 'pivview/piv_challenge1_E/piv_parameters.par'

    @staticmethod
    def get_plane_directory() -> pathlib.Path:
        """Return the path to the respective example PIV plane"""
        return user.TEST_DATA_DIRECTORY / 'pivview/piv_challenge1_E'

    @staticmethod
    def get_multiplane_directories() -> Tuple[pathlib.Path, pathlib.Path, pathlib.Path]:
        """Copies the piv_challenge1_E data to three directories in the tmp directory
        Two planes have three nc files, one plane has 2 nc files only"""
        try:
            from netCDF4 import Dataset as ncDataset
        except ImportError:
            raise ImportError('Package netCDF4 is not installed.')

        def _set_z_in_nc(nc_filename, z_val):
            with ncDataset(nc_filename, 'r+') as nc:
                nc.setncattr('origin_offset_z', z_val)
                for k, v in nc.variables.items():
                    if 'coord_min' in nc[k].ncattrs():
                        coord_min = nc[k].getncattr('coord_min')
                        coord_min[-1] = z_val
                        nc[k].setncattr('coord_min', coord_min)
                        coord_max = nc[k].getncattr('coord_max')
                        coord_max[-1] = z_val
                        nc[k].setncattr('coord_max', coord_max)

        src_dir = user.TEST_DATA_DIRECTORY / 'pivview/piv_challenge1_E'
        nc_files = sorted(src_dir.glob('*[0-9].nc'))
        if len(nc_files) == 0:
            raise FileNotFoundError('No nc files found in piv_challenge1_E test data directory. Contact the developer.')

        plane0 = generate_temporary_directory(prefix='mplane/')
        _ = shutil.copy2(src_dir / 'piv_parameters.par', plane0.joinpath('piv_parameter.par'))
        dst = shutil.copy2(nc_files[0], plane0.joinpath('f0.nc'))
        _set_z_in_nc(dst, -5.)
        dst = shutil.copy2(nc_files[1], plane0.joinpath('f1.nc'))
        _set_z_in_nc(dst, -5.)
        dst = shutil.copy2(nc_files[2], plane0.joinpath('f2.nc'))
        _set_z_in_nc(dst, -5.)

        plane1 = generate_temporary_directory(prefix='mplane/')
        _ = shutil.copy2(src_dir / 'piv_parameters.par', plane1.joinpath('piv_parameter.par'))
        dst = shutil.copy2(nc_files[3], plane1.joinpath('f0.nc'))
        _set_z_in_nc(dst, 0.)
        dst = shutil.copy2(nc_files[4], plane1.joinpath('f1.nc'))
        _set_z_in_nc(dst, 0.)
        dst = shutil.copy2(nc_files[5], plane1.joinpath('f2.nc'))
        _set_z_in_nc(dst, 0.)

        plane2 = generate_temporary_directory(prefix='mplane/')
        _ = shutil.copy2(src_dir / 'piv_parameters.par', plane2.joinpath('piv_parameter.par'))
        dst = shutil.copy2(nc_files[6], plane2.joinpath('f0.nc'))
        _set_z_in_nc(dst, 10.)
        dst = shutil.copy2(nc_files[7], plane2.joinpath('f1.nc'))
        _set_z_in_nc(dst, 10.)

        return plane0, plane1, plane2

    @staticmethod
    def get_snapshot_nc_files():
        """Return a list of sorted nc files"""
        files = sorted((user.TEST_DATA_DIRECTORY / "pivview/piv_challenge1_E").glob('E00A*.nc'))
        if len(files) == 0:
            raise FileNotFoundError('No nc files found in piv_challenge1_E test data directory. Contact the developer.')
        return files

    @staticmethod
    def get_avg_file() -> pathlib.Path:
        """Return the path to the avg.dat file"""
        file = user.TEST_DATA_DIRECTORY / "pivview/piv_challenge1_E/avg.dat"
        if not file.exists():
            raise FileNotFoundError(
                'No avg.dat file found in piv_challenge1_E test data directory. Contact the developer.')
        return file

    @staticmethod
    def get_rms_file() -> pathlib.Path:
        """Return the path to the rms.dat file"""
        file = user.TEST_DATA_DIRECTORY / "pivview/piv_challenge1_E/rms.dat"
        if not file.exists():
            raise FileNotFoundError(
                'No rms.dat file found in piv_challenge1_E test data directory. Contact the developer.')
        return file

    @staticmethod
    def get_reyn_file() -> pathlib.Path:
        """Return the path to the reyn.dat file"""
        file = user.TEST_DATA_DIRECTORY / "pivview/piv_challenge1_E/reyn.dat"
        if not file.exists():
            raise FileNotFoundError(
                'No reyn.dat file found in piv_challenge1_E test data directory. Contact the developer.')
        return file


class OpenPIV:
    """OpenPIV tutorial class"""

    @staticmethod
    def get_snapshot_txt_file() -> pathlib.Path:
        """Return snapshot piv result from ILA vortex"""
        return user.TEST_DATA_DIRECTORY / "openpiv" / "vortex.txt"

    @staticmethod
    def get_parameter_file() -> pathlib.Path:
        """Return openpiv parameters as file"""
        return user.TEST_DATA_DIRECTORY / "openpiv" / "openpiv.par"

    @staticmethod
    def get_plane_directory() -> pathlib.Path:
        """Return the path to the respective example PIV plane"""
        plane = generate_temporary_directory()
        snapshot_filename = OpenPIV.get_snapshot_txt_file()
        parameter_filename = OpenPIV.get_parameter_file()
        shutil.copy2(snapshot_filename, plane / 'vortex1.txt')
        shutil.copy2(snapshot_filename, plane / 'vortex2.txt')
        shutil.copy2(snapshot_filename, plane / 'vortex3.txt')
        shutil.copy2(snapshot_filename, plane / 'vortex4.txt')
        shutil.copy2(parameter_filename, plane / parameter_filename.name)
        return plane

    @staticmethod
    def get_multiplane_directories() -> Tuple[pathlib.Path, pathlib.Path, pathlib.Path]:
        case = generate_temporary_directory()
        plane1 = case / 'p1'
        plane2 = case / 'p2'
        plane3 = case / 'p3'
        plane1.mkdir()
        plane2.mkdir()
        plane3.mkdir()
        snapshot_filename = OpenPIV.get_snapshot_txt_file()
        parameter_filename = OpenPIV.get_parameter_file()
        shutil.copy2(snapshot_filename, plane1 / 'vortex1.txt')
        shutil.copy2(snapshot_filename, plane1 / 'vortex2.txt')
        shutil.copy2(snapshot_filename, plane1 / 'vortex3.txt')
        shutil.copy2(snapshot_filename, plane2 / 'vortex1.txt')
        shutil.copy2(snapshot_filename, plane2 / 'vortex2.txt')
        shutil.copy2(snapshot_filename, plane2 / 'vortex3.txt')
        shutil.copy2(snapshot_filename, plane3 / 'vortex1.txt')
        shutil.copy2(snapshot_filename, plane3 / 'vortex2.txt')
        shutil.copy2(parameter_filename, plane1 / parameter_filename.name)
        shutil.copy2(parameter_filename, plane2 / parameter_filename.name)
        shutil.copy2(parameter_filename, plane3 / parameter_filename.name)
        return plane1, plane2, plane3


class Davis:
    """LaVision Davis tutorial data class"""

    @staticmethod
    def get_set_file() -> pathlib.Path:
        """Return the path to the davis set file"""
        return user.TEST_DATA_DIRECTORY / 'davis/dtau04.0.set'

    @staticmethod
    def get_vc7_files() -> List[pathlib.Path]:
        """Return the path to the davis set file"""
        return sorted((user.TEST_DATA_DIRECTORY / 'davis/dtau04.0/').glob('*.vc7'))



import json
import pathlib
import unittest

import piv2hdf

__this_dir__ = pathlib.Path(__file__).parent


class TestVersion(unittest.TestCase):

    def test_version(self):
        this_version = 'x.x.x'
        setupcfg_filename = __this_dir__ / '../setup.cfg'
        with open(setupcfg_filename, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'version =' in line:
                    this_version = line.split(' = ')[-1].strip()
        self.assertEqual(piv2hdf.__version__, this_version)

    def test_codemeta(self):
        """checking if the version in codemeta.json is the same as the one of the piv2hdf"""

        with open(__this_dir__ / '../codemeta.json', 'r') as f:
            codemeta = json.loads(f.read())

        self.assertEqual(codemeta['version'], piv2hdf.__version__)

    def test_citation_cff(self):
        """checking if the version in CITATION.cff is the same as the one of the piv2hdf"""
        this_version = 'x.x.x'
        setupcfg_filename = __this_dir__ / '../CITATION.cff'
        with open(setupcfg_filename, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'version: ' in line:
                    this_version = line.split(':')[-1].strip()
        self.assertEqual(piv2hdf.__version__, this_version)



import h5py
import h5rdmtoolbox as h5tbx


def compute_time_averages(h5, **kwargs):
    """Compute time averages of the velocity field.
    This is only reasonable if one dimension scale is relative time!
    """
    # print('computing time averages is called!')
    if not isinstance(h5, h5py.File):
        with h5tbx.File(h5, 'r+') as h5:
            return compute_time_averages(h5, **kwargs)

    for sn in ('x_velocity', 'y_velocity', 'z_velocity'):
        lazy_u_ds = h5.find_one({'standard_name': sn})
        if lazy_u_ds is None:
            # print(f'Could not find dataset with standard_name "{sn}"')
            continue

        u_ds = h5[lazy_u_ds.name]
        dim_ds_standard_names = [dim[0].attrs.get('standard_name', '') for dim in u_ds.dims]
        if 'relative_time' not in dim_ds_standard_names:
            return  # skip this post-processing function, there is no relative_time dimension attached!

        u_mean_ds = u_ds[()].mean(dim='reltime')

        mean_name = f'mean_{u_ds.basename}'
        if mean_name in h5:
            # print(f'Dataset with name "{mean_name}" already exists. Skipping...')
            continue
        u_ds.parent.create_dataset(mean_name,
                                   data=u_mean_ds,
                                   attrs={'units': u_ds.attrs['units'],
                                          'standard_name': f'arithmetic_mean_of_{sn}'})
